from .EMUtils import omega, k, VTEMFun, TriangleFun, SineFun
from .AnalyticUtils import (
    MagneticDipoleFields,
    MagneticDipoleVectorPotential,
    MagneticLoopVectorPotential,
    orientationDict,
)
from .CurrentUtils import (
    getSourceTermLineCurrentPolygon,
    getStraightLineCurrentIntegral,
)
